/** Pikku esimerkki tuottaja-kuluttaja -tilanteesta ilman
 *  synkronointia. Ei tietenkään toimi toivotulla tavalla, koska
 *  säikeiden vuoronnus on sattumanvaraista. Katso korjattu koodi:
 *  tuottaja_kuluttaja.c
 *
 *  HUOM: Esimerkkien lyhentämiseksi näissä ei tarkisteta,
 *  onnistuivatko operaatiot. Oikeassa ohjelmassa AINA tarkistettava
 *  esim. onnistuiko säikeen luonti vai ei, ja käsiteltävä
 *  epäonnistuminen!!
 */
#define _POSIX_C_SOURCE 200809L
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <ctype.h>

#define BUF_SIZE 20

/* Meillä on nyt leluesimerkissä vain merkkejä, mutta vois olla
 * videoframeja tai muuta isompaa. Idea olisi sama silloinkin.
 */
typedef char data_t;

typedef struct {
    int iluku;
    int ikirjoitus;
    data_t data[BUF_SIZE];
} rengaspuskuri_t;

void pikkuinen_tauko(int nsec) {
    struct timespec tim;
    tim.tv_sec = 0;
    tim.tv_nsec = nsec;
    nanosleep(&tim, NULL);
}

void rengas_alusta(rengaspuskuri_t *p) {
    p->iluku = 0;
    p->ikirjoitus = 0;
}

void rengas_kirjoita(rengaspuskuri_t *p, data_t merkki) {
    p->data[p->ikirjoitus] = merkki;
    p->ikirjoitus = (p->ikirjoitus + 1) % BUF_SIZE;
}

data_t rengas_lue(rengaspuskuri_t *p) {
    data_t palautettava = p->data[p->iluku];
    p->iluku = (p->iluku + 1) % BUF_SIZE;
    return palautettava;
}

void* tuottaja(void *arg)
{
    char c;
    static int i=0;
    const char tviittausta[] = "Tviit...";

    rengaspuskuri_t *p = (rengaspuskuri_t*) arg;

    for (int n=0; n<1000; n++) {
        rengas_kirjoita(p, tviittausta[i]);
        i = (i + 1) % 8;
        pikkuinen_tauko(5000+(rand()%1500));
    }

    return NULL;
}

void* kuluttaja(void *arg)
{
    char c;
    rengaspuskuri_t *p = (rengaspuskuri_t*) arg;

    for (int n=0;n<1000;n++) {
        c = rengas_lue(p);
        fputc(toupper(c), stdout);
        fflush(stdout);
        pikkuinen_tauko(80000+(rand()%1500));
    }
    return NULL;
}

int main(int argc, char **argv)
{
    rengaspuskuri_t rengas;

    pthread_t t,k;

    rengas_alusta(&rengas);

    pthread_create(&t, NULL, tuottaja, &rengas);
    pthread_create(&k, NULL, kuluttaja, &rengas);

    pthread_join(t, NULL);
    pthread_join(k, NULL);

    return 0;
}
